==> disasm-0.6.tar.bz2: disasm-0.6/view_elf64.cc <==
class ver_tag {
public:
  const char *file, *ver;
  ver_tag() {}
  ver_tag( const char *file0, const char *ver0 ): file(file0), ver(ver0) {}
};

void objtbl_view_dyntab() {
  Elf *elf = elf_begin( fd, ELF_C_READ, 0 );
  if ( !elf ) return;

  Elf_Scn *scn = 0, *dynsym_scn = 0, *gnuver_scn = 0, *gnuverd_scn = 0,
	  *gnuverr_scn = 0;
  Elf64_Shdr *shdr = 0;

  for ( scn = elf_nextscn( elf, scn ); scn && (!dynsym_scn || !gnuver_scn || !gnuverd_scn || !gnuverr_scn); scn = elf_nextscn( elf, scn ) ) {

  	shdr = elf64_getshdr( scn );
	if ( shdr->sh_type == SHT_DYNSYM ) dynsym_scn = scn;
	else if ( shdr->sh_type == SHT_GNU_versym ) gnuver_scn = scn;
	else if ( shdr->sh_type == SHT_GNU_verdef ) gnuverd_scn = scn;
	else if ( shdr->sh_type == SHT_GNU_verneed ) gnuverr_scn = scn;
  }

  push_keyhints();
  clear_keyhints();

  WINDOW *w = newwin( 21, 61, (LINES-21)/2, (COLS-61)/2 );
  win_stack.push_back( w );

  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(6) );
  mvwprintw( w, 0, 2, "Dynamic Symbols Table" );
  mvwaddstr( w, 1, 1, "Symbol name                 Version        File" );

  select_list_ent sle;
  vector <select_list_ent> lst;

  if ( dynsym_scn ) {
  	map <int, ver_tag> ver_str;

	if ( gnuverd_scn ) {
		Elf_Data *verd_data = elf_getdata( gnuverd_scn, 0 );
		shdr = elf64_getshdr( gnuverd_scn );
		for ( int offs = 0; offs < shdr->sh_size; ) {
			Elf64_Verdef *vd = (Elf64_Verdef*)((char*)verd_data->d_buf + offs);
			if ( vd->vd_cnt ) {
				Elf64_Verdaux *vdx = (Elf64_Verdaux*)((char*)vd + vd->vd_aux);
				ver_str[vd->vd_ndx] = ver_tag( 0, elf_strptr( elf, shdr->sh_link, vdx->vda_name ) );
			}

			if ( vd->vd_next ) offs += vd->vd_next; else break;
		}
	}

	if ( gnuverr_scn ) {
		Elf_Data *verr_data = elf_getdata( gnuverr_scn, 0 );
		shdr = elf64_getshdr( gnuverr_scn );
		for ( int offs = 0; offs < shdr->sh_size; ) {
			Elf64_Verneed *vn = (Elf64_Verneed*)((char*)verr_data->d_buf + offs);
			char *file = elf_strptr( elf, shdr->sh_link, vn->vn_file );
			char *v_offs = (char*)vn + vn->vn_aux;
			for ( int i = 0; i < vn->vn_cnt; i++ ) {
				Elf64_Vernaux *v = (Elf64_Vernaux*)v_offs;
				ver_str[v->vna_other] = ver_tag( file, elf_strptr( elf, shdr->sh_link, v->vna_name ) );
				v_offs += v->vna_next;
			}

			if ( vn->vn_next ) offs += vn->vn_next; else break;
		}
	}

	Elf_Data *sym_data = elf_getdata( dynsym_scn, 0 );
	Elf_Data *ver_data = gnuver_scn? elf_getdata( gnuver_scn, 0 ): 0;

  	shdr = elf64_getshdr( dynsym_scn );
	int nsym = shdr->sh_size / shdr->sh_entsize;
	int dynstr_section_index = shdr->sh_link;

	for ( int i = 0; i < nsym; i++ ) {
	  	char *symname = elf_strptr( elf, dynstr_section_index,
			((Elf64_Sym*)sym_data->d_buf)[i].st_name );

		int veridx = -1;
		if ( ver_data )
			veridx = ((Elf64_Half*)ver_data->d_buf)[i] & ~0x8000;

		const char *ver = "<undefined>";
		char vernum[10], *file = 0;

		if ( veridx == 0 ) ver = "<local>";
		else if ( veridx == 1 ) ver = "<exported>";
		else {
			map <int, ver_tag>::iterator it = ver_str.find( veridx );
			if ( it != ver_str.end() ) {
				ver = (char*)it->second.ver;
				file = (char*)it->second.file;
			}
			else {
				sprintf( vernum, "<%d>", veridx );
				ver = vernum;
			}
		}

		sle.x = 1;
		sle.y = i + 2;
		asprintf( &sle.str, "%-27.27s %-14.14s %-16.16s",
			symname[0]? symname: "<undefined>", ver, file? file: "" );
		lst.push_back( sle );
	}
  }
  
  do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), lst );
  
  delwin( w );
  pop_winstack();
  pop_keyhints();

  elf_end( elf );
}

void objtbl_view_vertags( Elf *elf, int link_scn, Elf64_Verneed *vn ) {
  push_keyhints();
  clear_keyhints();

  WINDOW *w = newwin( 18, 62, (LINES-18)/2, (COLS-62)/2 );
  win_stack.push_back( w );

  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(6) );
  mvwprintw( w, 0, 2, "Version Tags" );
  mvwaddstr( w, 1, 1, "Offset   Hash     Flgs Indx NameIndx Tag            NextOffs" );

  select_list_ent sle;
  vector <select_list_ent> lst;

  char *v_offs = (char*)vn + vn->vn_aux;
  for ( int i = 0; i < vn->vn_cnt; i++ ) {
	Elf64_Vernaux *v = (Elf64_Vernaux*)v_offs;

	char *name = elf_strptr( elf, link_scn, v->vna_name );

	sle.x = 1;
	sle.y = i + 2;
	asprintf( &sle.str, "%08x %08x %04x %04x %08x %-14.14s %08x",
		(uint32_t)(v_offs - (char*)vn), v->vna_hash, v->vna_flags, v->vna_other,
		v->vna_name, name, v->vna_next );
	lst.push_back( sle );

	v_offs += v->vna_next;
  }
  
  do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), lst );
  
  delwin( w );
  pop_winstack();
  pop_keyhints();
}

class ver_tag2 {
public:
  Elf *elf;
  int link_scn;
  Elf64_Verneed *vn;

  ver_tag2( Elf *elf0, int link_scn0, Elf64_Verneed *vn0 ): elf(elf0), link_scn(link_scn0), vn(vn0) {}
};

void objtbl_vers_keyhandler( int c, void *data ) {
  if ( c == KEY_F(3) ) {
	ver_tag2 *vt2 = (ver_tag2*)data;
	objtbl_view_vertags( vt2->elf, vt2->link_scn, vt2->vn );
  }
}

void objtbl_view_verobj() {
  Elf *elf = elf_begin( fd, ELF_C_READ, 0 );
  if ( !elf ) return;

  Elf_Scn *scn = 0, *gnuverr_scn = 0;
  Elf64_Shdr *shdr = 0;

  for ( scn = elf_nextscn( elf, scn ); scn; scn = elf_nextscn( elf, scn ) ) {

  	shdr = elf64_getshdr( scn );
	if ( shdr->sh_type == SHT_GNU_verneed ) {
		gnuverr_scn = scn;
		break;
	}
  }

  push_keyhints();
  clear_keyhints();
  set_keyhint( 3, "View" );

  WINDOW *w = newwin( 21, 68, (LINES-21)/2, (COLS-68)/2 );
  win_stack.push_back( w );

  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(6) );
  mvwprintw( w, 0, 2, "Versioned Objects" );
  mvwaddstr( w, 1, 1, "Offset   Vers TCnt NameIndx Filename             TagOffs  NextOffs" );

  select_list_ent sle;
  vector <select_list_ent> lst;

  if ( gnuverr_scn ) {
	Elf_Data *verr_data = elf_getdata( gnuverr_scn, 0 );
	shdr = elf64_getshdr( gnuverr_scn );
	int i = 0;
	for ( int offs = 0; offs < shdr->sh_size; ) {
		Elf64_Verneed *vn = (Elf64_Verneed*)((char*)verr_data->d_buf + offs);
		char *file = elf_strptr( elf, shdr->sh_link, vn->vn_file );

		sle.x = 1;
		sle.y = (i++) + 2;
		asprintf( &sle.str, "%08x %04x %04x %08x %-20.20s %08x %08x",
			offs, vn->vn_version, vn->vn_cnt, vn->vn_file, file,
			vn->vn_aux, vn->vn_next );
		sle.data = new ver_tag2( elf, shdr->sh_link, vn );
		lst.push_back( sle );

		if ( vn->vn_next ) offs += vn->vn_next; else break;
	}
  }
  
  do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), lst, objtbl_vers_keyhandler );

  for ( int i = 0; i < lst.size(); i++ )
	delete (ver_tag2*)lst[i].data;
  
  delwin( w );
  pop_winstack();
  pop_keyhints();

  elf_end( elf );
}

void objtbl_view_verdtags( Elf *elf, int link_scn, Elf64_Verdef *vd ) {
  push_keyhints();
  clear_keyhints();

  WINDOW *w = newwin( 12, 43, (LINES-12)/2, (COLS-43)/2 );
  win_stack.push_back( w );

  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(6) );
  mvwprintw( w, 0, 2, "Version Tags" );
  mvwaddstr( w, 1, 1, "Offset   NameIndx Tag            NextOffs" );

  select_list_ent sle;
  vector <select_list_ent> lst;

  char *v_offs = (char*)vd + vd->vd_aux;
  for ( int i = 0; i < vd->vd_cnt; i++ ) {
	Elf64_Verdaux *v = (Elf64_Verdaux*)v_offs;

	char *name = elf_strptr( elf, link_scn, v->vda_name );

	sle.x = 1;
	sle.y = i + 2;
	asprintf( &sle.str, "%08x %08x %-14.14s %08x",
		(uint32_t)(v_offs - (char*)vd), v->vda_name, name, v->vda_next );
	lst.push_back( sle );

	v_offs += v->vda_next;
  }
  
  do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), lst );
  
  delwin( w );
  pop_winstack();
  pop_keyhints();
}

class ver_tag3 {
public:
  Elf *elf;
  int link_scn;
  Elf64_Verdef *vd;

  ver_tag3( Elf *elf0, int link_scn0, Elf64_Verdef *vd0 ): elf(elf0), link_scn(link_scn0), vd(vd0) {}
};

void objtbl_versd_keyhandler( int c, void *data ) {
  if ( c == KEY_F(3) ) {
	ver_tag3 *vt3 = (ver_tag3*)data;
	objtbl_view_verdtags( vt3->elf, vt3->link_scn, vt3->vd );
  }
}

void objtbl_view_verdobj() {
  Elf *elf = elf_begin( fd, ELF_C_READ, 0 );
  if ( !elf ) return;

  Elf_Scn *scn = 0, *gnuverd_scn = 0;
  Elf64_Shdr *shdr = 0;

  for ( scn = elf_nextscn( elf, scn ); scn; scn = elf_nextscn( elf, scn ) ) {

  	shdr = elf64_getshdr( scn );
	if ( shdr->sh_type == SHT_GNU_verdef ) {
		gnuverd_scn = scn;
		break;
	}
  }

  push_keyhints();
  clear_keyhints();
  set_keyhint( 3, "View" );

  WINDOW *w = newwin( 21, 72, (LINES-21)/2, (COLS-72)/2 );
  win_stack.push_back( w );

  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(6) );
  mvwprintw( w, 0, 2, "Version Definitions" );
  mvwaddstr( w, 1, 1, "Offset   Vers Flgs Indx Cnt  Hash     AuxOffs  Tag            NextOffs" );

  select_list_ent sle;
  vector <select_list_ent> lst;

  if ( gnuverd_scn ) {
	Elf_Data *verd_data = elf_getdata( gnuverd_scn, 0 );
	shdr = elf64_getshdr( gnuverd_scn );
	int i = 0;
	for ( int offs = 0; offs < shdr->sh_size; ) {
		Elf64_Verdef *vd = (Elf64_Verdef*)((char*)verd_data->d_buf + offs);

		char *name = 0;
		if ( vd->vd_cnt ) {
			Elf64_Verdaux *vdx = (Elf64_Verdaux*)((char*)vd + vd->vd_aux);
			name = elf_strptr( elf, shdr->sh_link, vdx->vda_name );
		}

		sle.x = 1;
		sle.y = (i++) + 2;
		asprintf( &sle.str, "%08x %04x %04x %04x %04x %08x %08x %-14.14s %08x",
			offs, vd->vd_version, vd->vd_flags, vd->vd_ndx,
			vd->vd_cnt, vd->vd_hash, vd->vd_aux, name? name: "",
			vd->vd_next );
		sle.data = new ver_tag3( elf, shdr->sh_link, vd );
		lst.push_back( sle );

		if ( vd->vd_next ) offs += vd->vd_next; else break;
	}
  }
  
  do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), lst, objtbl_versd_keyhandler );

  for ( int i = 0; i < lst.size(); i++ )
	delete (ver_tag3*)lst[i].data;
  
  delwin( w );
  pop_winstack();
  pop_keyhints();

  elf_end( elf );
}

int view_elf_objtbl() {
  WINDOW *w;
  int i, c, rc, offs;
  unsigned char *p;
  char *str;
  select_list_ent sle;
  vector <select_list_ent> lst;
  Elf *elf;
  Elf64_Ehdr *ehdr;
  Elf_Scn *scn;
  Elf64_Shdr *shdr;

  push_keyhints();
  clear_keyhints();
  set_keyhint( 2, "SavDat" );
  set_keyhint( 6, "SymTbl" );
  set_keyhint( 7, "ImpVer" );
  set_keyhint( 8, "ExpVer" );

  w = newwin( 18, 71, (LINES-18)/2, (COLS-71)/2 );
  win_stack.push_back( w );

  keypad( w, TRUE );
  box( w, ACS_VLINE, ACS_HLINE );
  wbkgd( w, COLOR_PAIR(6) );
  mvwprintw( w, 0, 2, "Object Table" );
  mvwaddstr( w, 1, 1, "Section name   Indx Type     Address  FileOffs Size     Link Flags   " );

  elf = elf_begin( fd, ELF_C_READ, 0 );
  ehdr = elf64_getehdr( elf );
  scn = 0;

  for ( i = 0;; i++ ) {
  	scn = elf_nextscn( elf, scn );
	if ( !scn ) break;
	shdr = elf64_getshdr( scn );
	str = elf_strptr( elf, ehdr->e_shstrndx, shdr->sh_name );
  	sle.x = 1;
	sle.y = i + 2;
	asprintf( &sle.str, "%-14.14s %04x %08x %08x %08x %08x %04x %08x",
                str, (i + 1) & 0xffff, shdr->sh_type, (uint32_t)shdr->sh_addr,
		(uint32_t)shdr->sh_offset, (uint32_t)shdr->sh_size, shdr->sh_link & 0xffff,
		(uint32_t)shdr->sh_flags );
	sle.data = scn;
	lst.push_back( sle );
  }
 
  rc = do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), lst, objtbl_keyhandler );
  offs = -1;

  if ( rc != -1 ) {
	shdr = elf64_getshdr( (Elf_Scn*)lst[rc].data );
	offs = shdr->sh_offset;
  }
  elf_end( elf );

  delwin( w );
  pop_winstack();
  pop_keyhints();

  return offs;
}

==> disasm-0.8.tar.bz2: disasm-0.8/view_elf64.cc <==
class ver_tag {
public:
  const char *file, *ver;
  ver_tag() {}
  ver_tag( const char *file0, const char *ver0 ): file(file0), ver(ver0) {}
};

static sel_list_persist dyn_pers;

void objtbl_view_dyntab() {
  Elf *elf = elf_begin( fd, ELF_C_READ, 0 );
  if ( !elf ) return;

  Elf_Scn *scn = 0, *dynsym_scn = 0, *gnuver_scn = 0, *gnuverd_scn = 0,
	  *gnuverr_scn = 0;
  Elf64_Shdr *shdr = 0;

  for ( scn = elf_nextscn( elf, scn ); scn && (!dynsym_scn || !gnuver_scn || !gnuverd_scn || !gnuverr_scn); scn = elf_nextscn( elf, scn ) ) {

  	shdr = elf64_getshdr( scn );
	if ( shdr->sh_type == SHT_DYNSYM ) dynsym_scn = scn;
	else if ( shdr->sh_type == SHT_GNU_versym ) gnuver_scn = scn;
	else if ( shdr->sh_type == SHT_GNU_verdef ) gnuverd_scn = scn;
	else if ( shdr->sh_type == SHT_GNU_verneed ) gnuverr_scn = scn;
  }

  push_keyhints();
  clear_keyhints();

  set_keyhint(6, "SortNa");
  set_keyhint(7, "SortVe");
  set_keyhint(8, "SortFi");

  WINDOW *w = newwin( 21, 61, (LINES-21)/2, (COLS-61)/2 );
  win_stack.push_back( w );

  keypad( w, TRUE );
  wbkgd( w, COLOR_PAIR(6) );
  box( w, ACS_VLINE, ACS_HLINE );
  mvwprintw( w, 0, 2, "Dynamic Symbols Table" );
  mvwhline ( w, 1, 1, ' ', 61 - 2 );
  mvwaddstr( w, 1, 1, "Symbol name                 Version        File" );

  select_list_ent sle;
  vector <select_list_ent> lst;

  if ( dynsym_scn ) {
  	map <int, ver_tag> ver_str;

	if ( gnuverd_scn ) {
		Elf_Data *verd_data = elf_getdata( gnuverd_scn, 0 );
		shdr = elf64_getshdr( gnuverd_scn );
		for ( int offs = 0; offs < shdr->sh_size; ) {
			Elf64_Verdef *vd = (Elf64_Verdef*)((char*)verd_data->d_buf + offs);
			if ( vd->vd_cnt ) {
				Elf64_Verdaux *vdx = (Elf64_Verdaux*)((char*)vd + vd->vd_aux);
				ver_str[vd->vd_ndx] = ver_tag( 0, elf_strptr( elf, shdr->sh_link, vdx->vda_name ) );
			}

			if ( vd->vd_next ) offs += vd->vd_next; else break;
		}
	}

	if ( gnuverr_scn ) {
		Elf_Data *verr_data = elf_getdata( gnuverr_scn, 0 );
		shdr = elf64_getshdr( gnuverr_scn );
		for ( int offs = 0; offs < shdr->sh_size; ) {
			Elf64_Verneed *vn = (Elf64_Verneed*)((char*)verr_data->d_buf + offs);
			char *file = elf_strptr( elf, shdr->sh_link, vn->vn_file );
			char *v_offs = (char*)vn + vn->vn_aux;
			for ( int i = 0; i < vn->vn_cnt; i++ ) {
				Elf64_Vernaux *v = (Elf64_Vernaux*)v_offs;
				ver_str[v->vna_other] = ver_tag( file, elf_strptr( elf, shdr->sh_link, v->vna_name ) );
				v_offs += v->vna_next;
			}

			if ( vn->vn_next ) offs += vn->vn_next; else break;
		}
	}

	Elf_Data *sym_data = elf_getdata( dynsym_scn, 0 );
	Elf_Data *ver_data = gnuver_scn? elf_getdata( gnuver_scn, 0 ): 0;

  	shdr = elf64_getshdr( dynsym_scn );
	int nsym = shdr->sh_size / shdr->sh_entsize;
	int dynstr_section_index = shdr->sh_link;

	for ( int i = 0; i < nsym; i++ ) {
	  	char *symname = elf_strptr( elf, dynstr_section_index,
			((Elf64_Sym*)sym_data->d_buf)[i].st_name );

		int veridx = -1;
		if ( ver_data )
			veridx = ((Elf64_Half*)ver_data->d_buf)[i] & ~0x8000;

		const char *ver = "<undefined>";
		char vernum[10], *file = 0;

		if ( veridx == 0 ) ver = "<local>";
		else if ( veridx == 1 ) ver = "<exported>";
		else if ( veridx == -1 ) ;
		else {
			map <int, ver_tag>::iterator it = ver_str.find( veridx );
			if ( it != ver_str.end() ) {
				ver = (char*)it->second.ver;
				file = (char*)it->second.file;
			}
			else {
				sprintf( vernum, "<%d>", veridx );
				ver = vernum;
			}
		}

		sle.x = 1;
		sle.y = i + 2;
		asprintf( &sle.str, "%-27.27s %-14.14s %-16.16s",
			symname[0]? symname: "<undefined>", ver, file? file: "" );
		lst.push_back( sle );
	}
  }
  
  do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), &dyn_pers, lst,
		[&lst](int c, void *data) { return objtbl_dyntab_keyhandler( c, data, lst ); } );
  
  delwin( w );
  pop_winstack();
  pop_keyhints();

  elf_end( elf );
}

static sel_list_persist verreq_pers;

void objtbl_view_vertags( Elf *elf, int link_scn, Elf64_Verneed *vn ) {
  push_keyhints();
  clear_keyhints();

  WINDOW *w = newwin( 18, 62, (LINES-18)/2, (COLS-62)/2 );
  win_stack.push_back( w );

  keypad( w, TRUE );
  wbkgd( w, COLOR_PAIR(6) );
  box( w, ACS_VLINE, ACS_HLINE );
  mvwprintw( w, 0, 2, "Version Tags" );
  mvwaddstr( w, 1, 1, "Offset   Hash     Flgs Indx NameIndx Tag            NextOffs" );

  select_list_ent sle;
  vector <select_list_ent> lst;

  char *v_offs = (char*)vn + vn->vn_aux;
  for ( int i = 0; i < vn->vn_cnt; i++ ) {
	Elf64_Vernaux *v = (Elf64_Vernaux*)v_offs;

	char *name = elf_strptr( elf, link_scn, v->vna_name );

	sle.x = 1;
	sle.y = i + 2;
	asprintf( &sle.str, "%08x %08x %04x %04x %08x %-14.14s %08x",
		(uint32_t)(v_offs - (char*)vn), v->vna_hash, v->vna_flags, v->vna_other,
		v->vna_name, name, v->vna_next );
	lst.push_back( sle );

	v_offs += v->vna_next;
  }
  
  do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), &verreq_pers, lst );
  
  delwin( w );
  pop_winstack();
  pop_keyhints();
}

class ver_tag2 {
public:
  Elf *elf;
  int link_scn;
  Elf64_Verneed *vn;

  ver_tag2( Elf *elf0, int link_scn0, Elf64_Verneed *vn0 ): elf(elf0), link_scn(link_scn0), vn(vn0) {}
};

int objtbl_vers_keyhandler( int c, void *data ) {
  if ( c == KEY_F(3) ) {
	ver_tag2 *vt2 = (ver_tag2*)data;
	objtbl_view_vertags( vt2->elf, vt2->link_scn, vt2->vn );
  }
  return KEYH_ACTION_NONE;
}

static sel_list_persist verobj_pers;

void objtbl_view_verobj() {
  Elf *elf = elf_begin( fd, ELF_C_READ, 0 );
  if ( !elf ) return;

  Elf_Scn *scn = 0, *gnuverr_scn = 0;
  Elf64_Shdr *shdr = 0;

  for ( scn = elf_nextscn( elf, scn ); scn; scn = elf_nextscn( elf, scn ) ) {

  	shdr = elf64_getshdr( scn );
	if ( shdr->sh_type == SHT_GNU_verneed ) {
		gnuverr_scn = scn;
		break;
	}
  }

  push_keyhints();
  clear_keyhints();
  set_keyhint( 3, "View" );

  WINDOW *w = newwin( 21, 68, (LINES-21)/2, (COLS-68)/2 );
  win_stack.push_back( w );

  keypad( w, TRUE );
  wbkgd( w, COLOR_PAIR(6) );
  box( w, ACS_VLINE, ACS_HLINE );
  mvwprintw( w, 0, 2, "Versioned Objects" );
  mvwaddstr( w, 1, 1, "Offset   Vers TCnt NameIndx Filename             TagOffs  NextOffs" );

  select_list_ent sle;
  vector <select_list_ent> lst;

  if ( gnuverr_scn ) {
	Elf_Data *verr_data = elf_getdata( gnuverr_scn, 0 );
	shdr = elf64_getshdr( gnuverr_scn );
	int i = 0;
	for ( int offs = 0; offs < shdr->sh_size; ) {
		Elf64_Verneed *vn = (Elf64_Verneed*)((char*)verr_data->d_buf + offs);
		char *file = elf_strptr( elf, shdr->sh_link, vn->vn_file );

		sle.x = 1;
		sle.y = (i++) + 2;
		asprintf( &sle.str, "%08x %04x %04x %08x %-20.20s %08x %08x",
			offs, vn->vn_version, vn->vn_cnt, vn->vn_file, file,
			vn->vn_aux, vn->vn_next );
		sle.data = new ver_tag2( elf, shdr->sh_link, vn );
		lst.push_back( sle );

		if ( vn->vn_next ) offs += vn->vn_next; else break;
	}
  }
  
  do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), &verobj_pers, lst, objtbl_vers_keyhandler );

  for ( int i = 0; i < lst.size(); i++ )
	delete (ver_tag2*)lst[i].data;
  
  delwin( w );
  pop_winstack();
  pop_keyhints();

  elf_end( elf );
}

static sel_list_persist verdefaux_pers;

void objtbl_view_verdtags( Elf *elf, int link_scn, Elf64_Verdef *vd ) {
  push_keyhints();
  clear_keyhints();

  WINDOW *w = newwin( 12, 43, (LINES-12)/2, (COLS-43)/2 );
  win_stack.push_back( w );

  keypad( w, TRUE );
  wbkgd( w, COLOR_PAIR(6) );
  box( w, ACS_VLINE, ACS_HLINE );
  mvwprintw( w, 0, 2, "Version Tags" );
  mvwaddstr( w, 1, 1, "Offset   NameIndx Tag            NextOffs" );

  select_list_ent sle;
  vector <select_list_ent> lst;

  char *v_offs = (char*)vd + vd->vd_aux;
  for ( int i = 0; i < vd->vd_cnt; i++ ) {
	Elf64_Verdaux *v = (Elf64_Verdaux*)v_offs;

	char *name = elf_strptr( elf, link_scn, v->vda_name );

	sle.x = 1;
	sle.y = i + 2;
	asprintf( &sle.str, "%08x %08x %-14.14s %08x",
		(uint32_t)(v_offs - (char*)vd), v->vda_name, name, v->vda_next );
	lst.push_back( sle );

	v_offs += v->vda_next;
  }
  
  do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), &verdefaux_pers, lst );
  
  delwin( w );
  pop_winstack();
  pop_keyhints();
}

class ver_tag3 {
public:
  Elf *elf;
  int link_scn;
  Elf64_Verdef *vd;

  ver_tag3( Elf *elf0, int link_scn0, Elf64_Verdef *vd0 ): elf(elf0), link_scn(link_scn0), vd(vd0) {}
};

int objtbl_versd_keyhandler( int c, void *data ) {
  if ( c == KEY_F(3) ) {
	ver_tag3 *vt3 = (ver_tag3*)data;
	objtbl_view_verdtags( vt3->elf, vt3->link_scn, vt3->vd );
  }
  return KEYH_ACTION_NONE;
}

static sel_list_persist verdef_pers;

void objtbl_view_verdobj() {
  Elf *elf = elf_begin( fd, ELF_C_READ, 0 );
  if ( !elf ) return;

  Elf_Scn *scn = 0, *gnuverd_scn = 0;
  Elf64_Shdr *shdr = 0;

  for ( scn = elf_nextscn( elf, scn ); scn; scn = elf_nextscn( elf, scn ) ) {

  	shdr = elf64_getshdr( scn );
	if ( shdr->sh_type == SHT_GNU_verdef ) {
		gnuverd_scn = scn;
		break;
	}
  }

  push_keyhints();
  clear_keyhints();
  set_keyhint( 3, "View" );

  WINDOW *w = newwin( 21, 72, (LINES-21)/2, (COLS-72)/2 );
  win_stack.push_back( w );

  keypad( w, TRUE );
  wbkgd( w, COLOR_PAIR(6) );
  box( w, ACS_VLINE, ACS_HLINE );
  mvwprintw( w, 0, 2, "Version Definitions" );
  mvwaddstr( w, 1, 1, "Offset   Vers Flgs Indx Cnt  Hash     AuxOffs  Tag            NextOffs" );

  select_list_ent sle;
  vector <select_list_ent> lst;

  if ( gnuverd_scn ) {
	Elf_Data *verd_data = elf_getdata( gnuverd_scn, 0 );
	shdr = elf64_getshdr( gnuverd_scn );
	int i = 0;
	for ( int offs = 0; offs < shdr->sh_size; ) {
		Elf64_Verdef *vd = (Elf64_Verdef*)((char*)verd_data->d_buf + offs);

		char *name = 0;
		if ( vd->vd_cnt ) {
			Elf64_Verdaux *vdx = (Elf64_Verdaux*)((char*)vd + vd->vd_aux);
			name = elf_strptr( elf, shdr->sh_link, vdx->vda_name );
		}

		sle.x = 1;
		sle.y = (i++) + 2;
		asprintf( &sle.str, "%08x %04x %04x %04x %04x %08x %08x %-14.14s %08x",
			offs, vd->vd_version, vd->vd_flags, vd->vd_ndx,
			vd->vd_cnt, vd->vd_hash, vd->vd_aux, name? name: "",
			vd->vd_next );
		sle.data = new ver_tag3( elf, shdr->sh_link, vd );
		lst.push_back( sle );

		if ( vd->vd_next ) offs += vd->vd_next; else break;
	}
  }
  
  do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), &verdef_pers, lst, objtbl_versd_keyhandler );

  for ( int i = 0; i < lst.size(); i++ )
	delete (ver_tag3*)lst[i].data;
  
  delwin( w );
  pop_winstack();
  pop_keyhints();

  elf_end( elf );
}

static sel_list_persist obj_pers;

int view_elf_objtbl() {
  WINDOW *w;
  int i, c, rc, offs;
  unsigned char *p;
  char *str;
  select_list_ent sle;
  vector <select_list_ent> lst;
  Elf *elf;
  Elf64_Ehdr *ehdr;
  Elf_Scn *scn;
  Elf64_Shdr *shdr;

  push_keyhints();
  clear_keyhints();
  set_keyhint( 2, "SavDat" );
  set_keyhint( 4, "SortNa" );
  set_keyhint( 5, "SortAd" );
  set_keyhint( 6, "SymTbl" );
  set_keyhint( 7, "ImpVer" );
  set_keyhint( 8, "ExpVer" );

  w = newwin( 18, 71, (LINES-18)/2, (COLS-71)/2 );
  win_stack.push_back( w );

  keypad( w, TRUE );
  wbkgd( w, COLOR_PAIR(6) );
  box( w, ACS_VLINE, ACS_HLINE );
  mvwprintw( w, 0, 2, "Object Table" );
  mvwaddstr( w, 1, 1, "Section name   Indx Type     Address  FileOffs Size     Link Flags   " );

  elf = elf_begin( fd, ELF_C_READ, 0 );
  ehdr = elf64_getehdr( elf );
  scn = 0;

  for ( i = 0;; i++ ) {
  	scn = elf_nextscn( elf, scn );
	if ( !scn ) break;
	shdr = elf64_getshdr( scn );
	str = elf_strptr( elf, ehdr->e_shstrndx, shdr->sh_name );
  	sle.x = 1;
	sle.y = i + 2;
	asprintf( &sle.str, "%-14.14s %04x %08x %08x %08x %08x %04x %08x",
                str, (i + 1) & 0xffff, shdr->sh_type, (uint32_t)shdr->sh_addr,
		(uint32_t)shdr->sh_offset, (uint32_t)shdr->sh_size, shdr->sh_link & 0xffff,
		(uint32_t)shdr->sh_flags );
	sle.data = scn;
	lst.push_back( sle );
  }
 
  rc = do_list_selection( w, COLOR_PAIR(3), COLOR_PAIR(6), &dyn_pers, lst,
		[&lst](int c, void *data) { return objtbl_keyhandler( c, data, lst ); } );
  offs = -1;

  if ( rc != -1 ) {
	shdr = elf64_getshdr( (Elf_Scn*)lst[rc].data );
	offs = shdr->sh_offset;
  }
  elf_end( elf );

  delwin( w );
  pop_winstack();
  pop_keyhints();

  return offs;
}
